#line 1 "/fzi/prost/stone/SOS3-2/src/agg/List.c"
/* --------------------------------------------------------------------------
 * Copyright 1992 by Forschungszentrum Informatik (FZI)
 *
 * You can use and distribute this software under the terms of the licence
 * you should have received along with this program.
 * If not or if you want additional information, write to
 * Forschungszentrum Informatik, "STONE", Haid-und-Neu-Strasse 10-14,
 * D-7500 Karlsruhe 1, Germany.
 * --------------------------------------------------------------------------
 */
// **************************************************************************
// Module List                      03/07/89           Bernhard Schiefer (bs)
//                                                   modified : 24/10/91 (bs)
// **************************************************************************
// implements methods of classes: List, sos_List_node
// **************************************************************************

#include "sys.h"
#include "agg_err.h"
#include "trc_agg.h"

#include "agg_sos.h"

// **************************************************************************
void _sos_Object_List::local_initialize (sos_Object_List list)
// **************************************************************************
{  T_PROC ("sos_Object_List::local_initialize");
   TT (agg_H, T_ENTER);

   list.set_first (sos_List_node::make (NO_OBJECT));
   list.set_last (sos_List_node::make (NO_OBJECT));
   list.set_cardinality (0);

   TT (agg_H, T_LEAVE);
} // ** local_initialize **

// **************************************************************************
void _sos_Object_List::local_finalize (sos_Object_List list)
// **************************************************************************
{  T_PROC ("sos_Object_List::local_finalize");
   TT (agg_H, T_ENTER);

   list.clear();

   TT (agg_H, T_LEAVE);
} // ** local_finalize **

// **************************************************************************
void _sos_Object_List::append (sos_Typed_id &_tpid,sos_Object o)
// **************************************************************************
// append an element
{
   T_PROC ("sos_Object_List::append");
   TT (agg_H, T_ENTER);

   sos_List_node cn;
   cn = sos_List_node::create (sos_Object_List::make(_tpid,this).container());
   cn.set_val (o);

   if (sos_Object_List::make(_tpid,this).get_first() == NO_OBJECT)          // insert the first node
   {  sos_Object_List::make(_tpid,this).set_first (cn);
      sos_Object_List::make(_tpid,this).set_last  (cn);
      cn.set_succ (sos_List_node::make (NO_OBJECT));
      cn.set_pred (sos_List_node::make (NO_OBJECT));
   }
   else
   {  sos_List_node n = sos_Object_List::make(_tpid,this).get_last();
      n.insert_after (cn); 
      sos_Object_List::make(_tpid,this).set_last (cn);
   }

   sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() + 1);

   TT (agg_H, T_LEAVE);
} // ** append **

// **************************************************************************
void _sos_Object_List::insert (sos_Typed_id &_tpid,Index  position, sos_Object o)
// **************************************************************************
// insert an element at the specified position
// if position 0 is specified insert after the last element
{
   T_PROC ("sos_Object_List::insert");
   TT (agg_H, T_ENTER; TI (position));

   err_assert (position >= 0, "sos_Object_List::insert");

   if (position > sos_Object_List::make(_tpid,this).get_cardinality())
      position = 0;        // insert after last!

   if (position == 0)
      sos_Object_List::make(_tpid,this).append (o);
   else
   {  sos_List_node first_node = sos_Object_List::make(_tpid,this).get_first();
      sos_List_node cn = sos_List_node::create (sos_Object_List::make(_tpid,this).container());
      cn.set_val (o);
      if (first_node == NO_OBJECT)              // insert the first node 
      {  sos_Object_List::make(_tpid,this).set_first (cn);
         sos_Object_List::make(_tpid,this).set_last (cn);
	 cn.set_succ (sos_List_node::make (NO_OBJECT));
	 cn.set_pred (sos_List_node::make (NO_OBJECT));
      }
      else
      {  sos_List_node n = first_node.succ (position - 1);
         n.insert_before (cn);
         if (position == 1) sos_Object_List::make(_tpid,this).set_first (cn);
      } 
      sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() + 1);
   }
   TT (agg_H, T_LEAVE);

} // ** insert **

// **************************************************************************
void _sos_Object_List::remove (sos_Typed_id &_tpid,Index  position)
// **************************************************************************
{
   T_PROC ("sos_Object_List::remove");
   TT (agg_H, T_ENTER);

   sos_List_node first_node = sos_Object_List::make(_tpid,this).get_first();
   err_assert (first_node != NO_OBJECT, "sos_Object_List::remove");

   sos_List_node cn = first_node.succ (position - 1);
   err_assert (cn != NO_OBJECT, "sos_Object_List::remove");

#ifdef ATT
   if (cn.operator==(first_node))
      sos_Object_List::make(_tpid,this).set_first (first_node.succ());
   if (cn.operator==(sos_Object_List::make(_tpid,this).get_last()))
      sos_Object_List::make(_tpid,this).set_last (sos_Object_List::make(_tpid,this).get_last().pred());
#else
   if (cn == first_node)
      sos_Object_List::make(_tpid,this).set_first (first_node.succ());
   if (cn == sos_Object_List::make(_tpid,this).get_last())
      sos_Object_List::make(_tpid,this).set_last (sos_Object_List::make(_tpid,this).get_last().pred());
#endif

   cn.remove(); 
   sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() - 1);

   TT (agg_H, T_LEAVE);

} // ** remove **

// **************************************************************************
sos_Object _sos_Object_List::get_nth (sos_Typed_id &_tpid,Index pos)
// **************************************************************************
{
   T_PROC ("sos_Object_List::get_nth");
   TT (agg_H, T_ENTER; TI (pos));
    
   sos_List_node cn = sos_Object_List::make(_tpid,this).get_first();
   err_assert (cn != NO_OBJECT, "sos_Object_List::get_nth");

   cn = cn.succ (pos - 1);

   err_assert (cn != NO_OBJECT, "sos_Object_List::get_nth");

   sos_Object o = cn.get_val();

   TT (agg_H, T_LEAVE);
   return o;
} // ** get_nth **

// **************************************************************************
void _sos_Object_List::set_nth (sos_Typed_id &_tpid,Index pos, sos_Object o)
// **************************************************************************
// the value of the sos_Object at Index position is replaced by o
{
   T_PROC ("sos_Object_List::set_nth");
   TT (agg_H, T_ENTER; TI (pos));

   sos_List_node cn = sos_Object_List::make(_tpid,this).get_first();
   err_assert (cn != NO_OBJECT, "sos_Object_List::set_nth");

   cn = cn.succ (pos - 1);
   err_assert (cn != NO_OBJECT, "sos_Object_List::set_nth");

   cn.set_val (o);

   TT (agg_H, T_LEAVE);

} // ** set_nth **

// **************************************************************************
void _sos_Object_List::__plus_assign (sos_Typed_id &_tpid,sos_Object_List alist)
// **************************************************************************
{
   T_PROC ("sos_Object_List::operator+=");
   TT (agg_H, T_ENTER);

   for (sos_List_node cn = alist.get_first();
        (cn != NO_OBJECT);
        (cn = cn.succ()))
      sos_Object_List::make(_tpid,this).append (cn.get_val());

   TT (agg_H, T_LEAVE);
} // ** operator+= **

// **************************************************************************
Index _sos_Object_List::current_pos (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::current_pos");
   TT (agg_H, T_ENTER);

   Index result = 0;

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::current_pos");

   for (sos_List_node cn = sos_List_node::make (c.get_current());
        (cn != NO_OBJECT);
        (cn = cn.pred()))
      result++;

   TT (agg_H, T_LEAVE; TI (result));
   return result;
} // ** current_pos **

// **************************************************************************
sos_Bool _sos_Object_List::move_cursor (sos_Typed_id &_tpid,sos_Cursor c, Index position)
// **************************************************************************
{
   T_PROC ("sos_Object_List::move_cursor");
   TT (agg_H, T_ENTER; TI (position));

   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::move_cursor");

   sos_Bool success = TRUE;
   sos_List_node current = sos_Object_List::make(_tpid,this).get_first();

   if (current == NO_OBJECT)
      success = FALSE;
   else
   {  current = current.succ (position - 1);
      if (current != NO_OBJECT)
         c.set_current (current);
      else
	 success = FALSE;
   } // else

   TT (agg_H, T_LEAVE; TB(success));
   return success;
} // ** move_cursor **

// **************************************************************************
void _sos_Object_List::insert_before (sos_Typed_id &_tpid,sos_Cursor c, sos_Object o)
// **************************************************************************
{
   T_PROC ("sos_Object_List::insert_before");
   TT (agg_H, T_ENTER);

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::insert_before");

   sos_List_node cn = sos_List_node::create (sos_Object_List::make(_tpid,this).container());
   cn.set_val (o);
   sos_List_node::make (c.get_current()).insert_before (cn);
   sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() + 1);
   if (cn.pred() == NO_OBJECT)
      sos_Object_List::make(_tpid,this).set_first (cn);

   TT (agg_H, T_LEAVE);
}

// **************************************************************************
void _sos_Object_List::insert_after (sos_Typed_id &_tpid,sos_Cursor c, sos_Object o)
// **************************************************************************
{
   T_PROC ("sos_Object_List::insert_after");
   TT (agg_H, T_ENTER);

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::insert_after");

   sos_List_node cn = sos_List_node::create (sos_Object_List::make(_tpid,this).container());
   cn.set_val (o);
   sos_List_node::make (c.get_current()).insert_after (cn);
   sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() + 1);
   if (cn.succ() == NO_OBJECT)
      sos_Object_List::make(_tpid,this).set_last (cn);

   TT (agg_H, T_LEAVE);
}

// **************************************************************************
void _sos_Object_List::local_assign (sos_Object_List x, sos_Object o)
// **************************************************************************
{
   T_PROC ("sos_Object_List::local_assign");
   TT (agg_H, T_ENTER );

   sos_Object_List y = sos_Object_List::make (o);
   x.clear();
   x.operator+= (y);

   TT (agg_H, T_LEAVE);
} // ** local_assign **

// **************************************************************************
sos_Bool _sos_Object_List::local_equal (sos_Object_List x,
				       sos_Object      o,
				       sos_Eq_kind     eq_kind)
// **************************************************************************
{
   T_PROC ("sos_Object_List::local_equal");
   TT (agg_H, T_ENTER );

   sos_Bool result;

   if ((eq_kind EQ EQ_STRONG AND NOT o.has_type (x.type())) OR
       (eq_kind EQ EQ_WEAK   AND NOT o.isa     (x.type())))
      result = FALSE;
   else
   {  sos_Object_List y = sos_Object_List::make (o);
      if (x.get_cardinality() != y.get_cardinality())
	 result = FALSE;
      else
      {  result = TRUE;
	 sos_Bool based_on_equal = x.get_based_on_equal();
	 sos_Int  comp = 0;
	 agg_iterate_double (x, sos_Object x_e, y, sos_Object y_e, comp)
	 {  result = agg_same_entity (x_e, y_e, based_on_equal, eq_kind);
	    if (NOT result) break;
	 }
	 agg_iterate_double_end (x, x_e, y, y_e, comp);
	 result = (sos_Bool)(result AND comp == 0);
      }
   }

   TT (agg_H, T_LEAVE; TB(result));

   return result;
} // ** local_equal **

// **************************************************************************
sos_Int _sos_Object_List::local_hash_value (sos_Object_List x)
// **************************************************************************
{
   T_PROC ("sos_Object_List::local_hash_value");
   TT (agg_H, T_ENTER );

   sos_Int result;

   if (x.is_empty())
      result = 0;
   else
      result = x.get_nth (1).hash_value();

   TT (agg_H, T_LEAVE; TB(result));

   return result;
} // ** local_hash_value **

// **************************************************************************
sos_Object _sos_Object_List::get (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::get (sos_Cursor)");
   TT (agg_H, T_ENTER);

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::get (sos_Cursor)");

   sos_Object o = sos_List_node::make (c.get_current()).get_val();

   TT (agg_H, T_LEAVE);
   return o;

} // ** get **
 
// **************************************************************************
void _sos_Object_List::set (sos_Typed_id &_tpid,sos_Cursor c, sos_Object o)
// **************************************************************************
{
   T_PROC ("sos_Object_List::set");
   TT (agg_H, T_ENTER);

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::set");

   sos_List_node::make (c.get_current()).set_val (o);

   TT (agg_H, T_LEAVE);

} // ** set **
 
// **************************************************************************
void _sos_Object_List::remove_at (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::remove_at");
   TT (agg_H, T_ENTER);

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::remove_at");

   sos_List_node n = sos_List_node::make (c.get_current());

#ifdef ATT
   if (n.operator==(sos_Object_List::make(_tpid,this).get_first()))
      sos_Object_List::make(_tpid,this).set_first (sos_Object_List::make(_tpid,this).get_first().succ());
   if (n.operator==(sos_Object_List::make(_tpid,this).get_last()))
      sos_Object_List::make(_tpid,this).set_last (sos_Object_List::make(_tpid,this).get_last().pred());
#else
   if (n == sos_Object_List::make(_tpid,this).get_first())
      sos_Object_List::make(_tpid,this).set_first (sos_Object_List::make(_tpid,this).get_first().succ());
   if (n == sos_Object_List::make(_tpid,this).get_last())
      sos_Object_List::make(_tpid,this).set_last (sos_Object_List::make(_tpid,this).get_last().pred());
#endif

   sos_Object_List::make(_tpid,this).to_succ (c);
   n.remove();
   sos_Object_List::make(_tpid,this).set_cardinality (sos_Object_List::make(_tpid,this).get_cardinality() - 1);

   TT (agg_H, T_LEAVE);
} // ** remove_at **

// **************************************************************************
sos_Int _sos_Object_List::card (sos_Typed_id &_tpid)
// **************************************************************************
{
   T_PROC ("sos_Object_List::card");
   TT (agg_H, T_ENTER);

   sos_Int crd = sos_Object_List::make(_tpid,this).get_cardinality();

   TT (agg_H, T_LEAVE);
   return crd;
} // ** card **

// **************************************************************************
sos_Cursor _sos_Object_List::open_cursor (sos_Typed_id &_tpid,sos_Container ct)
// **************************************************************************
{
   // creates a new cursor-object and positions it 
   // to the first element

   T_PROC ("sos_Object_List::open_cursor");
   TT (agg_H, T_ENTER);

   sos_Cursor new_c = sos_Cursor::create (ct, sos_Object_List::make(_tpid,this));
   new_c.set_current (sos_Object_List::make(_tpid,this).get_first());

   TT (agg_H, T_LEAVE);
   return new_c;
} // ** open_cursor **

// **************************************************************************
void _sos_Object_List::close_cursor (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   // The result of any operation using sos_Cursor c
   // is undefined after this operation.

   T_PROC ("sos_Object_List::close_cursor");
   TT (agg_H, T_ENTER);

   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::close_cursor");

   c.destroy();

   TT (agg_H, T_LEAVE);
} // ** close_cursor **

// **************************************************************************
sos_Cursor _sos_Object_List::duplicate (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   // creates a new cursor-object for the Aggregate and
   // positions it on the same element as c

   T_PROC ("sos_Object_List::duplicate");
   TT (agg_H, T_ENTER);

   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::duplicate");

   sos_Cursor new_c = sos_Cursor::create (c.container(), sos_Object_List::make(_tpid,this));
   new_c.set_current (c.get_current());

   TT (agg_H, T_LEAVE);
   return new_c;
} // ** duplicate **

// **************************************************************************
sos_Bool _sos_Object_List::is_valid (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::is_valid");
   TT (agg_H, T_ENTER);
 
   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::is_valid");

   sos_Bool valid = (c.get_current() != NO_OBJECT);

   TT (agg_H, T_LEAVE; TB(valid));
   return valid;
} // ** is_valid **
 
// **************************************************************************
sos_Bool _sos_Object_List::to_first (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::to_first");
   TT (agg_H, T_ENTER);
 
   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::to_first");

   sos_List_node first = sos_Object_List::make(_tpid,this).get_first();
   c.set_current (first);

   sos_Bool valid = (first != NO_OBJECT);

   TT (agg_H, T_LEAVE; TB(valid));
   return valid;
} // ** to_first **
 
// **************************************************************************
sos_Bool _sos_Object_List::to_last (sos_Typed_id &_tpid,sos_Cursor c)
// **************************************************************************
{
   T_PROC ("sos_Object_List::to_last");
   TT (agg_H, T_ENTER);
 
   err_assert (c.defined_for (sos_Object_List::make(_tpid,this)), "sos_Object_List::to_last");

   sos_List_node last = sos_Object_List::make(_tpid,this).get_last();
   c.set_current (last);

   sos_Bool valid = (last != NO_OBJECT);

   TT (agg_H, T_LEAVE; TB(valid));
   return valid;
} // ** to_last **

// **************************************************************************
sos_Bool _sos_Object_List::to_succ (sos_Typed_id &_tpid,sos_Cursor c, sos_Int i)
// **************************************************************************
{
   // Move the sos_Cursor to the i-th successing element
   // The function returns FALSE if no such element exists.

   T_PROC ("sos_Object_List::to_succ");
   TT (agg_H, T_ENTER; TI (i));

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::to_succ");

   sos_List_node succ = (sos_List_node::make (c.get_current())).succ (i);
   c.set_current (succ);

   sos_Bool valid = (succ != NO_OBJECT);

   TT (agg_H, T_LEAVE; TB(valid));
   return (valid);
} // ** to_succ **

// **************************************************************************
sos_Bool _sos_Object_List::to_pred (sos_Typed_id &_tpid,sos_Cursor c, sos_Int i)
// **************************************************************************
{
   // Move the sos_Cursor to the previous i-th element
   // The function returns FALSE if no such element exists.

   T_PROC ("sos_Object_List::to_pred");
   TT (agg_H, T_ENTER; TI (i));

   err_assert (sos_Object_List::make(_tpid,this).is_valid (c), "sos_Object_List::to_pred");

   sos_List_node pred = (sos_List_node::make (c.get_current())).pred (i);
   c.set_current (pred);

   sos_Bool valid = (pred != NO_OBJECT);

   TT (agg_H, T_LEAVE; TB(valid));
   return (valid);
} // ** to_pred **

// ***************************    sos_List_node    **************************

// **************************************************************************
void _sos_List_node::insert_before (sos_Typed_id &_tpid,sos_List_node n)
// **************************************************************************
{
   T_PROC ("sos_List_node::insert_before");
   TT (agg_H, T_ENTER);

   sos_List_node pred = sos_List_node::make(_tpid,this).get_pred();

   n.set_pred (pred);
   n.set_succ (sos_List_node::make(_tpid,this));
   sos_List_node::make(_tpid,this).set_pred (n);
   if (pred != NO_OBJECT) pred.set_succ (n);

   TT (agg_H, T_LEAVE);
}  // ** insert_before **

// **************************************************************************
void _sos_List_node::insert_after (sos_Typed_id &_tpid,sos_List_node n)
// **************************************************************************
{
   T_PROC ("sos_List_node::insert_after");
   TT (agg_H, T_ENTER);

   sos_List_node succ = sos_List_node::make(_tpid,this).get_succ();

   n.set_pred (sos_List_node::make(_tpid,this));
   n.set_succ (succ);
   sos_List_node::make(_tpid,this).set_succ (n);
   if (succ != NO_OBJECT) succ.set_pred (n);

   TT (agg_H, T_LEAVE);
}  // ** insert_after **

// **************************************************************************
void _sos_List_node::remove (sos_Typed_id &_tpid)
// **************************************************************************
{
   T_PROC ("sos_List_node::remove");
   TT (agg_H, T_ENTER);

   sos_List_node pred = sos_List_node::make(_tpid,this).get_pred();
   sos_List_node succ = sos_List_node::make(_tpid,this).get_succ();

   if (pred != NO_OBJECT) pred.set_succ (succ);
   if (succ != NO_OBJECT) succ.set_pred (pred);

   sos_List_node::make(_tpid,this).destroy();      // destroy the node

   TT (agg_H, T_LEAVE);
}  // ** remove **

// **************************************************************************
sos_List_node _sos_List_node::succ (sos_Typed_id &_tpid,sos_Int steps)
// **************************************************************************
{
   T_PROC ("sos_List_node::succ");
   TT (agg_H, T_ENTER; TI (steps));

   sos_List_node node = sos_List_node::make(_tpid,this);

   for (sos_Int i = 1;
        (i <= steps) AND (node != NO_OBJECT);
        i++)
       node = node.get_succ ();

   TT (agg_H, T_LEAVE; TI(i));
   return node;
}  // ** succ **

// **************************************************************************
sos_List_node _sos_List_node::pred (sos_Typed_id &_tpid,sos_Int steps)
// **************************************************************************
{
   T_PROC ("sos_List_node::pred");
   TT (agg_H, T_ENTER; TI (steps));

   sos_List_node node = sos_List_node::make(_tpid,this);

   for (sos_Int i = 1;
        (i <= steps) AND (node != NO_OBJECT);
        i++)
       node = node.get_pred();

   TT (agg_H, T_LEAVE; TI(i));
   return node;
}  // ** pred **
